part of mila;

class ModelFactory {
  ClassMirror _mirror;
  _ModelMirrorUtil _mirrorUtil;

  static Map<Type, ModelFactory> _cache = <Type, ModelFactory>{};

  factory ModelFactory(Type type) {
    if (!_cache.containsKey(type)) {
      _cache[type] = new ModelFactory._createNew(type);
    }
    return _cache[type];
  }

  ModelFactory._createNew(Type type) {
    _mirror = reflectClass(type);
    _mirrorUtil = new _ModelMirrorUtil(_mirror);
  }

  Model createEmpty() =>
      _mirror.newInstance(const Symbol(''), []).reflectee;

  Model _fillModel(Model model, Map attribute) {
    var reflection = reflect(model);

    attribute.keys.forEach((key) {
      var symbol = new Symbol(key),
          value = attribute[key],
          varFields = _mirrorUtil.varFields,
          modelVarFields = _mirrorUtil.modelVarFields;

      if (varFields.containsKey(symbol)) {
        if (modelVarFields.containsKey(symbol)) {
          var modelFactory = new ModelFactory(modelVarFields[symbol]
              .type
              .reflectedType);

          value = modelFactory.create(value);
        } else {
          if (_mirrorUtil.hasGeneric(symbol)) {
            var genericType = _mirrorUtil.getGenerics(symbol),
                lastMirror = genericType.removeLast();

            if (_mirrorUtil.hasModel(symbol)) {
              value = _traverseField(value, lastMirror);
            }
            new _TypeChecker(value, genericType).check();
          }
        }
      } else {
        throw 'no field named $key';
      }
      reflection.setField(symbol, value);
    });

    return reflection.reflectee;
  }

  _traverseField(data, ClassMirror classMirror) {
    if (data is Iterable) {
      return data.map((el) =>
          _traverseField(el, classMirror)).toList(growable: false);
    } else if (data is Map) {
      var dataKeys = data.keys.map((e) => new Symbol(e)),
          modelFields = new _ModelMirrorUtil(classMirror).fields,
          modelFieldSymbols = modelFields.keys,
          isModel = dataKeys.every((symbol) =>
              modelFieldSymbols.contains(symbol)),
          mappedValues;

      if (isModel || data.containsKey('id')) {
        var mFactory = new ModelFactory(classMirror.reflectedType);
        mappedValues = mFactory.create(data);

        return mappedValues;
      }
      mappedValues = data.values.map((el) =>
          _traverseField(el, classMirror));

      return new Map.fromIterables(data.keys, mappedValues);
    }
  }

  Model create(Map attribute) =>
      _fillModel(createEmpty(), attribute);
}