part of mila;

bool _isDynamic(TypeMirror typeMirror) =>
    currentMirrorSystem().dynamicType == typeMirror;

class _ModelMirrorUtil extends _ClassMirrorUtil {
  _ModelMirrorUtil(ClassMirror mirror) : super(mirror);

  bool _fieldChecker(DeclarationMirror dMirror) =>
      dMirror.metadata.any((meta) => meta.reflectee is _Field);

  List<InstanceMirror> getRule(Symbol symbol) =>
      varFields[symbol]
        .metadata
        .where((iMirror) => iMirror.reflectee is Rule)
        .toList(growable: false);

  bool hasGeneric(Symbol symbol) {
    var generic = varFields[symbol].type.typeArguments;

    return generic.isNotEmpty &&
        generic.last != currentMirrorSystem().dynamicType;
  }

  List<ClassMirror> getGenerics(Symbol symbol) {

    if (!fields.containsKey(symbol)) {
      throw 'no field with the following name: ${MirrorSystem.getName(symbol)}';
    }

    var declMirror = fields[symbol],
        current = declMirror.type,
        container = [current];

    while (current.typeArguments.isNotEmpty) {
      var generic = current.typeArguments.last;
      container.add(generic);
      current = generic;
    }

    return container;
  }

  bool hasModel(Symbol symbol) {
    var generics = getGenerics(symbol);

    if (generics.isEmpty) return false;

    var value = generics.last;

    if (value is! ClassMirror) {
      return false;
    }

    try {
      var instance = value.newInstance(const Symbol(''), []).reflectee;

      if (instance is Model) {
        return true;
      }
      return false;
    } catch (e) {
      return false;
    }
  }

  Map<Symbol, DeclarationMirror> get _fields {
    var  container = new Map(),
         variablesAsList = variables.values.where(_fieldChecker),
         getterAsList = getter.values.where(_fieldChecker),
         variablesAsMap = new Map.fromIterable(variablesAsList, 
             key: (dMirror) => dMirror.simpleName, 
             value: (dMirror) => dMirror),
         getterAsMap = new Map.fromIterable(getterAsList, 
             key: (dMirror) => dMirror.simpleName, 
             value: (dMirror) => dMirror);

    container
      ..addAll(variablesAsMap)
      ..addAll(getterAsMap);

    return container;
  }

  Map<Symbol, DeclarationMirror> get fields {
    var current = this,
        container = {};

    do {
      current._fields.forEach((key, value) =>
          container.putIfAbsent(key, () => value));

      current = new _ModelMirrorUtil(current.mirror.superclass);
    } while (current.mirror.reflectedType != Object);

    return container;
  }

  Map<Symbol, VariableMirror> get varFields {

    bool isVariable(DeclarationMirror dMirror) =>
        !dMirror.isPrivate && dMirror is VariableMirror;

    var asList = fields.values.where(isVariable),
        asMap = new Map.fromIterable(asList,
            key: (dMirror) => dMirror.simpleName,
            value: (dMirror) => dMirror);

    return asMap;
  }

  Map<Symbol, VariableMirror> get modelVarFields {

    bool isModel(VariableMirror vMirror) {
      ClassMirror classMirror = vMirror.type;

      try {
        var instance = classMirror
              .newInstance(const Symbol(''), [])
              .reflectee;

        return instance is Model;
      } catch (e) {
        return false;
      }
    }

    var asList = varFields.values.where(isModel),
        asMap = new Map.fromIterable(asList,
            key: (dMirror) => dMirror.simpleName,
            value: (dMirror) => dMirror);

    return asMap;
  }
}