part of mila.test;

const isfoo = const ValueIsFoo();
const required = const ValueIsRequired();

class ValueIsFoo extends Rule {

  const ValueIsFoo();

  validate(value) {
    if (value != 'foo') throw 'value is not foo';
  }
}

class ValueIsRequired extends Rule {

  const ValueIsRequired();

  validate(value) {
    if (value == null) throw 'value is required';
  }
}

class NameIsFoo extends Model {
  @field
  @isfoo
  String name;
}

class NestedRequired extends Model {
  @field
  @required
  List<Map<String, List<NameIsFoo>>> nested;
}

void validationTest() {
  group('validationTest', () {

    test('validate is foo', () {
      var nameIsFoo = new NameIsFoo()..name = 'foo';

      nameIsFoo
        .validate()
        .then(expectAsync1((val) {
          expect(val.name, equals('foo'));
        }));
    });

    test('validate is not foo', () {
      var nameIsFoo = new NameIsFoo()..name = 'not foo';

      nameIsFoo
        .validate()
        .catchError(expectAsync1((val) {
          expect(val, equals('value is not foo'));
        }));
    });

    test('nested required satisfied', () {
      var required = new NestedRequired()
            ..nested = [
              {'foo' : [new NameIsFoo()..name = 'foo',
                        new NameIsFoo()..name = 'foo']},
              {'bar' : [new NameIsFoo()..name = 'foo',
                        new NameIsFoo()..name = 'foo']}
            ];

      required
        .validate()
        .then(expectAsync1((value) {
          expect(required is NestedRequired, isTrue);
        }));
    });

    test('nested required unsatisfied', () {
      var required = new NestedRequired();

      required
        .validate()
        .catchError(expectAsync1((value) {
          expect(value, equals('value is required'));
        }));
    });

  });
}